==> simple-repository-2.4.0.tgz: package/lib/SimpleDynamoRepository.ts <==
import {
  DeleteCommand,
  DynamoDBDocumentClient,
  QueryCommand,
  ScanCommand,
  GetCommand,
  PutCommand,
  QueryCommandInput,
  ScanCommandInput,
  BatchWriteCommandInput,
  BatchWriteCommand,
} from '@aws-sdk/lib-dynamodb'
import { SimpleRepositoryNotFoundError } from './SimpleRepositoryNotFoundError'
import { DynamoDBServiceException } from '@aws-sdk/client-dynamodb'

export interface SimpleDocumentType {
  id: string
  createdAt: string
  updatedAt: string
}

type FindAllByQuery = {
  lastEvaluatedId?: string
  scanCommandInput: Omit<ScanCommandInput, 'TableName'>
}

// Dynamo is limited to batches of 25 items.
const MAX_BATCH_SIZE = 25

type PutRequest = { PutRequest: { Item: SimpleDocumentType } }
type DeleteRequest = { DeleteRequest: { Key: { id: string } } }
type WriteRequest = PutRequest | DeleteRequest

const DEFAULT_MAX_RETRIES = 10

// See https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Programming.Errors.html#Programming.Errors.MessagesAndCodes
const RETRYABLE_ERRORS = [
  'ProvisionedThroughputExceededException',
  'ThrottlingException',
  'ItemCollectionSizeLimitExceededException',
  'LimitExceededException',
  'ProvisionedThroughputExceeded',
  'RequestLimitExceeded',
  'UnrecognizedClientException',
]

export type BackOffProvider = (retries: number) => Promise<void>

const exponentialBackoff: BackOffProvider = async (retries: number): Promise<void> => {
  const ms = (2 ** retries - 1) * 700 + Math.random() * 600 + 50
  await new Promise((res) => setTimeout(res, ms))
}

export interface SimpleDynamoRepositoryOptions {
  backOffProvider?: BackOffProvider
  maxRetries?: number
  batchSize?: number
}

interface InternalBatchExecutionResult<RequestType extends WriteRequest> {
  unprocessed: RequestType[]
}

/**
 * The result of a batch execution.
 * Includes document ids of unprocessed items.
 */
export interface BatchExecutionResult {
  unprocessed: string[]
}

/**
 * A simple repository for DynamoDB
 * @template DocumentType the type of the document
 * example:
 * ```ts
 * type User {
 *  id: string
 *  name: string
 * }
 * class UserRepository extends SimpleDynamoRepository<User> {}
 * const userRepository = new UserRepository(new DynamoDBClient({ region: 'eu-central-1' }), 'users')
 * ```
 */
export abstract class SimpleDynamoRepository<DocumentType extends SimpleDocumentType> {
  protected readonly client: DynamoDBDocumentClient
  protected readonly tableName: string
  protected backOffProvider: BackOffProvider
  protected maxRetries: number
  protected batchSize: number

  constructor(
    client: DynamoDBDocumentClient,
    tableName: string,
    options?: SimpleDynamoRepositoryOptions
  ) {
    this.client = client
    this.tableName = tableName
    this.backOffProvider = options?.backOffProvider ?? exponentialBackoff
    this.maxRetries = options?.maxRetries ?? DEFAULT_MAX_RETRIES
    this.batchSize = Math.min(options?.batchSize ?? MAX_BATCH_SIZE, MAX_BATCH_SIZE)
  }

  /**
   * Find a document by id
   * @param id
   * @throws DocumentNotFoundError if the document does not exist
   * @returns the document
   */
  async findOne(id: string): Promise<DocumentType> {
    const doc = await this.client.send(new GetCommand({ Key: { id }, TableName: this.tableName }))

    if (!doc.Item) throw new SimpleRepositoryNotFoundError(id)

    return doc.Item as DocumentType
  }

  /**
   * Find a document by the global secondary index
   * @param key the name of the secondary index
   * @param value the value of the secondary index
   * @returns all documents with the given secondary index
   */
  async findBy(secondaryKey: string, value: string): Promise<DocumentType[]> {
    const queryCommandInput: QueryCommandInput = {
      TableName: this.tableName,
      IndexName: secondaryKey,
      KeyConditionExpression: '#key = :value',
      ExpressionAttributeNames: { '#key': secondaryKey },
      ExpressionAttributeValues: { ':value': value },
    }
    const docs = await this.client.send(new QueryCommand(queryCommandInput))

    return docs.Items as DocumentType[]
  }

  /**
   * Find the first document by the global secondary index
   * @param key the name of the secondary index
   * @param value the value of the secondary index
   * @returns the first document with the given secondary index
   */
  async findOneBy(secondaryKey: string, value: string): Promise<DocumentType> {
    const queryCommandInput: QueryCommandInput = {
      TableName: this.tableName,
      IndexName: secondaryKey,
      KeyConditionExpression: '#key = :value',
      ExpressionAttributeNames: { '#key': secondaryKey },
      ExpressionAttributeValues: { ':value': value },
    }
    const docs = await this.client.send(new QueryCommand(queryCommandInput))

    if (docs && docs.Items && docs.Items.length > 0) return docs.Items[0] as DocumentType

    throw new SimpleRepositoryNotFoundError()
  }

  /**
   * Save a document
   * @param doc
   * @returns the saved document
   */
  async save(doc: DocumentType): Promise<DocumentType> {
    await this.client.send(
      new PutCommand({
        Item: doc,
        TableName: this.tableName,
      })
    )

    return doc
  }

  /**
   * Save a list of documents in batches.
   * It will send batches of 25 items synchronously.
   * @param documents list of documents that will be saved in batches
   * @returns {BatchExecutionResult}
   */
  async batchSave(documents: DocumentType[]): Promise<BatchExecutionResult> {
    const queue: PutRequest[] = documents.map((doc) => ({ PutRequest: { Item: doc } }))

    const result = await this.executeBatchRequests<PutRequest>(queue)

    const toDocumentId = (req: PutRequest) => req.PutRequest.Item.id

    return { unprocessed: result.unprocessed.map(toDocumentId) }
  }

  /**
   * Delete a document by id
   * @param id
   * @throws DocumentNotFoundError if the document does not exist
   * @returns void
   */
  async delete(id: string): Promise<void> {
    try {
      await this.client.send(
        new DeleteCommand({
          Key: { id },
          TableName: this.tableName,
          ConditionExpression: 'attribute_exists(id)',
        })
      )
    } catch (error: any) {
      if (error.name === 'ConditionalCheckFailedException')
        throw new SimpleRepositoryNotFoundError(id)
      else throw error
    }
  }

  /**
   * Delete a list of documents by id in batches.
   * It will send batches of 25 items synchronously.
   * @param ids document ids that will be deleted in batches
   * @returns {BatchExecutionResult}
   */
  async batchDelete(ids: string[]): Promise<BatchExecutionResult> {
    const queue: DeleteRequest[] = ids.map((id) => ({ DeleteRequest: { Key: { id } } }))

    const result = await this.executeBatchRequests<DeleteRequest>(queue)

    const toDocumentId = (req: DeleteRequest) => req.DeleteRequest.Key.id

    return { unprocessed: result.unprocessed.map(toDocumentId) }
  }

  /**
   * Find all documents in the table
   * @param limit
   * @param lastEvaluatedId the id of the last document in the previous page
   * @returns all documents in the table
   * example:
   * ```ts
   * let lastEvaluatedId: string | undefined = undefined
   * do {
   *   const users: User[] = await userRepository.findAll(10, lastEvaluatedId)
   *   lastEvaluatedId = users[users.length - 1]?.id || undefined
   * } while (lastEvaluatedId)
   * ```
   */
  async findAll(limit = 1000, lastEvaluatedId?: string): Promise<DocumentType[]> {
    const scanCommandInput: ScanCommandInput = {
      TableName: this.tableName,
      Limit: limit,
    }

    if (lastEvaluatedId) scanCommandInput.ExclusiveStartKey = { id: lastEvaluatedId }

    const docs = await this.client.send(new ScanCommand(scanCommandInput))

    return docs.Items as DocumentType[]
  }

  /**
   * Find all documents in the table by query
   * @param query ScanCommandInput
   * @param lastEvaluatedId the id of the last document in the previous page
   * @returns all documents in the table
   * example:
   * ```ts
   * let lastEvaluatedId: string | undefined = undefined
   * do {
   *   const users: User[] = await userRepository.findAllBy(query, lastEvaluatedId)
   *   lastEvaluatedId = users[users.length - 1]?.id || undefined
   * } while (lastEvaluatedId)
   * ```
   */

  async findAllBy(query: FindAllByQuery): Promise<DocumentType[]> {
    const scanCommandInput: ScanCommandInput = {
      ...query.scanCommandInput,
      TableName: this.tableName,
    }

    if (query?.lastEvaluatedId) scanCommandInput.ExclusiveStartKey = { id: query.lastEvaluatedId }

    const docs = await this.client.send(new ScanCommand(scanCommandInput))

    return docs.Items as DocumentType[]
  }

  /**
   * Count all documents in the table
   * @returns the number of documents in the table
   */
  async count(): Promise<number> {
    const docs = await this.client.send(
      new ScanCommand({ TableName: this.tableName, Select: 'COUNT' })
    )

    return docs.Count || 0
  }

  /**
   * Find by query
   * @param queryCommandInput
   * @returns all documents that match the query
   */
  async queryBy(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType[]> {
    const docs = await this.client.send(
      new QueryCommand({
        ...queryCommandInput,
        TableName: this.tableName,
      })
    )

    return docs.Items as DocumentType[]
  }

  /**
   * Find one by query
   * @param queryCommandInput
   * @returns the first document that matches the query
   */
  async queryOneBy(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType> {
    const docs = await this.client.send(
      new QueryCommand({
        ...queryCommandInput,
        TableName: this.tableName,
      })
    )

    if (docs && docs.Items && docs.Items.length > 0) return docs.Items[0] as DocumentType

    throw new SimpleRepositoryNotFoundError()
  }

  protected async executeBatchRequests<RequestType extends WriteRequest>(
    queue: RequestType[]
  ): Promise<InternalBatchExecutionResult<RequestType>> {
    let retries = 0

    while (queue.length > 0) {
      const batch = queue.splice(0, Math.min(queue.length, this.batchSize))
      const input: BatchWriteCommandInput = { RequestItems: { [this.tableName]: batch } }

      let failedRequests: RequestType[] | undefined = []
      try {
        const result = await this.client.send(new BatchWriteCommand(input))
        if (result.UnprocessedItems) {
          failedRequests = result.UnprocessedItems?.[this.tableName] as RequestType[]
        }
      } catch (err) {
        if (this.isRetryableError(err)) failedRequests = batch
        else throw err
      }

      queue.unshift(...failedRequests)

      if (failedRequests?.length > 0) {
        if (retries >= this.maxRetries) return { unprocessed: queue }

        await this.backOffProvider(retries)

        retries++
      }
    }

    return { unprocessed: queue }
  }

  is5xxError(err: DynamoDBServiceException): boolean {
    const statusCode = err.$metadata?.httpStatusCode
    return statusCode != undefined && statusCode >= 500 && statusCode < 600
  }

  isRetryableError(err: unknown): boolean {
    return (
      err instanceof DynamoDBServiceException &&
      (this.is5xxError(err) || RETRYABLE_ERRORS.includes(err.name))
    )
  }
}

==> simple-repository-2.5.0.tgz: package/lib/SimpleDynamoRepository.ts <==
import {
  DeleteCommand,
  DynamoDBDocumentClient,
  QueryCommand,
  ScanCommand,
  GetCommand,
  PutCommand,
  QueryCommandInput,
  ScanCommandInput,
  BatchWriteCommandInput,
  BatchWriteCommand,
} from '@aws-sdk/lib-dynamodb'
import { SimpleRepositoryNotFoundError } from './SimpleRepositoryNotFoundError'
import { DynamoDBServiceException } from '@aws-sdk/client-dynamodb'

export interface SimpleDocumentType {
  id: string
  createdAt: string
  updatedAt: string
}

type FindAllByQuery = {
  lastEvaluatedId?: string
  scanCommandInput: Omit<ScanCommandInput, 'TableName'>
}

// Dynamo is limited to batches of 25 items.
const MAX_BATCH_SIZE = 25

type PutRequest = { PutRequest: { Item: SimpleDocumentType } }
type DeleteRequest = { DeleteRequest: { Key: { id: string } } }
type WriteRequest = PutRequest | DeleteRequest

const DEFAULT_MAX_RETRIES = 10

// See https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Programming.Errors.html#Programming.Errors.MessagesAndCodes
const RETRYABLE_ERRORS = [
  'ProvisionedThroughputExceededException',
  'ThrottlingException',
  'ItemCollectionSizeLimitExceededException',
  'LimitExceededException',
  'ProvisionedThroughputExceeded',
  'RequestLimitExceeded',
  'UnrecognizedClientException',
]

export type BackOffProvider = (retries: number) => Promise<void>

const exponentialBackoff: BackOffProvider = async (retries: number): Promise<void> => {
  const ms = (2 ** retries - 1) * 700 + Math.random() * 600 + 50
  await new Promise((res) => setTimeout(res, ms))
}

export interface SimpleDynamoRepositoryOptions {
  backOffProvider?: BackOffProvider
  maxRetries?: number
  batchSize?: number
}

interface InternalBatchExecutionResult<RequestType extends WriteRequest> {
  unprocessed: RequestType[]
}

/**
 * The result of a batch execution.
 * Includes document ids of unprocessed items.
 */
export interface BatchExecutionResult {
  unprocessed: string[]
}

/**
 * A simple repository for DynamoDB
 * @template DocumentType the type of the document
 * example:
 * ```ts
 * type User {
 *  id: string
 *  name: string
 * }
 * class UserRepository extends SimpleDynamoRepository<User> {}
 * const userRepository = new UserRepository(new DynamoDBClient({ region: 'eu-central-1' }), 'users')
 * ```
 */
export abstract class SimpleDynamoRepository<DocumentType extends SimpleDocumentType> {
  protected readonly client: DynamoDBDocumentClient
  protected readonly tableName: string
  protected backOffProvider: BackOffProvider
  protected maxRetries: number
  protected batchSize: number

  constructor(
    client: DynamoDBDocumentClient,
    tableName: string,
    options?: SimpleDynamoRepositoryOptions
  ) {
    this.client = client
    this.tableName = tableName
    this.backOffProvider = options?.backOffProvider ?? exponentialBackoff
    this.maxRetries = options?.maxRetries ?? DEFAULT_MAX_RETRIES
    this.batchSize = Math.min(options?.batchSize ?? MAX_BATCH_SIZE, MAX_BATCH_SIZE)
  }

  /**
   * Find a document by id
   * @param id
   * @throws DocumentNotFoundError if the document does not exist
   * @returns the document
   */
  async findOne(id: string): Promise<DocumentType> {
    const doc = await this.client.send(new GetCommand({ Key: { id }, TableName: this.tableName }))

    if (!doc.Item) throw new SimpleRepositoryNotFoundError(id)

    return doc.Item as DocumentType
  }

  /**
   * Find a document by the global secondary index
   * @param key the name of the secondary index
   * @param value the value of the secondary index
   * @returns all documents with the given secondary index
   */
  async findBy(secondaryKey: string, value: string): Promise<DocumentType[]> {
    const queryCommandInput: QueryCommandInput = {
      TableName: this.tableName,
      IndexName: secondaryKey,
      KeyConditionExpression: '#key = :value',
      ExpressionAttributeNames: { '#key': secondaryKey },
      ExpressionAttributeValues: { ':value': value },
    }
    const docs = await this.client.send(new QueryCommand(queryCommandInput))

    return docs.Items as DocumentType[]
  }

  /**
   * Find the first document by the global secondary index
   * @param key the name of the secondary index
   * @param value the value of the secondary index
   * @returns the first document with the given secondary index
   */
  async findOneBy(secondaryKey: string, value: string): Promise<DocumentType> {
    const queryCommandInput: QueryCommandInput = {
      TableName: this.tableName,
      IndexName: secondaryKey,
      KeyConditionExpression: '#key = :value',
      ExpressionAttributeNames: { '#key': secondaryKey },
      ExpressionAttributeValues: { ':value': value },
    }
    const docs = await this.client.send(new QueryCommand(queryCommandInput))

    if (docs && docs.Items && docs.Items.length > 0) return docs.Items[0] as DocumentType

    throw new SimpleRepositoryNotFoundError()
  }

  /**
   * Save a document
   * @param doc
   * @returns the saved document
   */
  async save(doc: DocumentType): Promise<DocumentType> {
    await this.client.send(
      new PutCommand({
        Item: doc,
        TableName: this.tableName,
      })
    )

    return doc
  }

  /**
   * Save a list of documents in batches.
   * It will send batches of 25 items synchronously.
   * @param documents list of documents that will be saved in batches
   * @returns {BatchExecutionResult}
   */
  async batchSave(documents: DocumentType[]): Promise<BatchExecutionResult> {
    const queue: PutRequest[] = documents.map((doc) => ({ PutRequest: { Item: doc } }))

    const result = await this.executeBatchRequests<PutRequest>(queue)

    const toDocumentId = (req: PutRequest) => req.PutRequest.Item.id

    return { unprocessed: result.unprocessed.map(toDocumentId) }
  }

  /**
   * Delete a document by id
   * @param id
   * @throws DocumentNotFoundError if the document does not exist
   * @returns void
   */
  async delete(id: string): Promise<void> {
    try {
      await this.client.send(
        new DeleteCommand({
          Key: { id },
          TableName: this.tableName,
          ConditionExpression: 'attribute_exists(id)',
        })
      )
    } catch (error: any) {
      if (error.name === 'ConditionalCheckFailedException')
        throw new SimpleRepositoryNotFoundError(id)
      else throw error
    }
  }

  /**
   * Delete a list of documents by id in batches.
   * It will send batches of 25 items synchronously.
   * @param ids document ids that will be deleted in batches
   * @returns {BatchExecutionResult}
   */
  async batchDelete(ids: string[]): Promise<BatchExecutionResult> {
    const queue: DeleteRequest[] = ids.map((id) => ({ DeleteRequest: { Key: { id } } }))

    const result = await this.executeBatchRequests<DeleteRequest>(queue)

    const toDocumentId = (req: DeleteRequest) => req.DeleteRequest.Key.id

    return { unprocessed: result.unprocessed.map(toDocumentId) }
  }

  /**
   * Find all documents in the table
   * @param limit
   * @param lastEvaluatedId the id of the last document in the previous page
   * @returns all documents in the table
   * example:
   * ```ts
   * let lastEvaluatedId: string | undefined = undefined
   * do {
   *   const users: User[] = await userRepository.findAll(10, lastEvaluatedId)
   *   lastEvaluatedId = users[users.length - 1]?.id || undefined
   * } while (lastEvaluatedId)
   * ```
   */
  async findAll(limit = 1000, lastEvaluatedId?: string): Promise<DocumentType[]> {
    const scanCommandInput: ScanCommandInput = {
      TableName: this.tableName,
      Limit: limit,
    }

    if (lastEvaluatedId) scanCommandInput.ExclusiveStartKey = { id: lastEvaluatedId }

    const docs = await this.client.send(new ScanCommand(scanCommandInput))

    return docs.Items as DocumentType[]
  }

  /**
   * Find all documents in the table by query
   * @param query ScanCommandInput
   * @param lastEvaluatedId the id of the last document in the previous page
   * @returns all documents in the table
   * example:
   * ```ts
   * let lastEvaluatedId: string | undefined = undefined
   * do {
   *   const users: User[] = await userRepository.findAllBy(query, lastEvaluatedId)
   *   lastEvaluatedId = users[users.length - 1]?.id || undefined
   * } while (lastEvaluatedId)
   * ```
   */

  async findAllBy(query: FindAllByQuery): Promise<DocumentType[]> {
    const scanCommandInput: ScanCommandInput = {
      ...query.scanCommandInput,
      TableName: this.tableName,
    }

    if (query?.lastEvaluatedId) scanCommandInput.ExclusiveStartKey = { id: query.lastEvaluatedId }

    const docs = await this.client.send(new ScanCommand(scanCommandInput))

    return docs.Items as DocumentType[]
  }

  /**
   * Count all documents in the table
   * @returns the number of documents in the table
   */
  async count(): Promise<number> {
    const docs = await this.client.send(
      new ScanCommand({ TableName: this.tableName, Select: 'COUNT' })
    )

    return docs.Count || 0
  }

  /**
   * Find by query
   * @param queryCommandInput
   * @returns all documents that match the query
   */
  async queryBy(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType[]> {
    const docs = await this.client.send(
      new QueryCommand({
        ...queryCommandInput,
        TableName: this.tableName,
      })
    )

    return docs.Items as DocumentType[]
  }

  /**
   * Find one by query
   * @param queryCommandInput
   * @returns the first document that matches the query
   */
  async queryOneBy(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType> {
    const docs = await this.client.send(
      new QueryCommand({
        ...queryCommandInput,
        TableName: this.tableName,
      })
    )

    if (docs && docs.Items && docs.Items.length > 0) return docs.Items[0] as DocumentType

    throw new SimpleRepositoryNotFoundError()
  }

  protected async executeBatchRequests<RequestType extends WriteRequest>(
    queue: RequestType[]
  ): Promise<InternalBatchExecutionResult<RequestType>> {
    let retries = 0

    while (queue.length > 0) {
      const batch = queue.splice(0, Math.min(queue.length, this.batchSize))
      const input: BatchWriteCommandInput = { RequestItems: { [this.tableName]: batch } }

      let failedRequests: RequestType[] | undefined = []
      try {
        const result = await this.client.send(new BatchWriteCommand(input))
        failedRequests = (result.UnprocessedItems?.[this.tableName] ?? []) as RequestType[]
      } catch (err) {
        if (this.isRetryableError(err)) failedRequests = batch
        else throw err
      }

      queue.unshift(...failedRequests)

      if (failedRequests?.length > 0) {
        if (retries >= this.maxRetries) return { unprocessed: queue }

        await this.backOffProvider(retries)

        retries++
      }
    }

    return { unprocessed: queue }
  }

  is5xxError(err: DynamoDBServiceException): boolean {
    const statusCode = err.$metadata?.httpStatusCode
    return statusCode != undefined && statusCode >= 500 && statusCode < 600
  }

  isRetryableError(err: unknown): boolean {
    return (
      err instanceof DynamoDBServiceException &&
      (this.is5xxError(err) || RETRYABLE_ERRORS.includes(err.name))
    )
  }
}
